#include <p33FJ64MC802.h>
//#include <p33fj12mc201.h> // other possible board, but only 1 QEI

// Configuration Bit Settings
_FOSCSEL(FNOSC_FRC) // use internal FRC
_FOSC(FCKSM_CSECMD & OSCIOFNC_ON) // Clock switching enabled and OSC2 pin is IO
_FWDT(FWDTEN_OFF) // watchdog timer controlled in software
_FPOR(FPWRT_PWR128) // power on reset timer value
_FICD(ICS_PGD3 & JTAGEN_OFF) // using PGD3 programming pins and no jtag

void init_SPI(void);
void write_SPI(short);
unsigned int QEI_read(int);
void QEI_enable(int, int);
void QEI_disable(int);
void QEI_reset(int);

int main() {

  // Configure Oscillator to operate Fcy at 40MHz
  // Fosc = Fin * M/(N1 * N2), Fcy = Fosc / 2
  // Fosc = 7.37*(43)/(2*2) = 79.2275MHz
  // So Fcy is 39.6ish MHz

  // Configure PLL prescaler, PLL postscaler, PLL divisor
  PLLFBD = 43;
  CLKDIVbits.PLLPOST = 0; // N2 = 2
  CLKDIVbits.PLLPRE = 0; // N1 = 2

  // Initiate Clock Switch to Internal FRC with PLL (NOSC = 0b001)
  __builtin_write_OSCCONH(0x01);
  __builtin_write_OSCCONL(0x01);

  // Wait for Clock switch to occur
  while (OSCCONbits.COSC != 0b001);

  // Wait for PLL to lock
  while (OSCCONbits.LOCK != 1) {
  };

  // unlock to set the pins
  __builtin_write_OSCCONL(OSCCON & (~(1 << 6))); // clear bit 6
  
  // pin select for QEI1
  RPINR14bits.QEB1R = 0b01000; // QEA1B Pin Select (RP8)
  RPINR14bits.QEA1R = 0b00111; // QEA1A Pin Select (RP7)
  RPINR15bits.INDX1R = 0b11111; // INDX1R (VSS)
  
  // pin select for QEI2
  //RPINR16bits.QEB2R = 0b00001; // QEA2B Pin Select (RP8)
  //RPINR16bits.QEA2R = 0b00010; // QEA2A Pin Select (RP7)
  //RPINR17bits.INDX2R = 0b11111; // INDX2R (VSS)

  // pin select for SPI1
  RPINR20bits.SCK1R = 0b01100; // SCK1 Pin Select (RP12)
  RPINR20bits.SDI1R = 0b01101; // SDI1 Pin Select (RP13)
  RPOR7bits.RP14R = 0b00111; // SDO1 Pin Select (RP14)
  RPINR21bits.SS1R = 0b01111; // SS1 Pin Select (RP15)

  // pin select for OC1
  RPOR1bits.RP3R = 0b10010; // OC1 on B3

  // lock pins
  __builtin_write_OSCCONL(OSCCON | (1 << 6)); // Set bit 6

  // turn off ADC pins
  AD1PCFGL = 0xFFFF;

  QEI_enable(1, 0xFFFF);
  //QEI_enable(2, 0xFFFF);
  QEI_reset(1);
  //QEI_reset(2);

  init_SPI();

  // Put a 50% 1kHz PWM on B3 to use as a heartbeat and know the board is working
  // Initialize Output Compare Module
  OC1CONbits.OCM = 0b000; // Disable Output Compare Module
  OC1R = 2500; // Write the duty cycle for the first PWM pulse
  OC1RS = 2500; // Write the duty cycle for the second PWM pulse
  OC1CONbits.OCTSEL = 0; // Select Timer 2 as output compare time base
  OC1CONbits.OCM = 0b110; // Select the Output Compare mode

  // Initialize and enable Timer2
  T2CONbits.TON = 0; // Disable Timer
  T2CONbits.TCS = 0; // Select internal instruction cycle clock
  T2CONbits.TGATE = 0; // Disable Gated Timer mode
  T2CONbits.TCKPS = 0b01; // Select 1:8 Prescaler
  TMR2 = 0x00; // Clear timer register
  PR2 = 5000; // Load the period value, so 40000000/8/5000 = 1000hz
  T2CONbits.TON = 1; // turn on timer2

  short read = 0;
  short pos;

  while (1) {
    // if the RX buffer is full
    if(SPI1STATbits.SPIRBF == 1) {
      read = SPI1BUF;

      // if got a 1, return the encoder value
      if (read == 1) {
        pos = QEI_read(1);
        SPI1BUF = pos;
        while (!SPI1STATbits.SPITBF); // wait for the data to be sent out
        //SPI1BUF = pos&0xFF;
        //while (!SPI1STATbits.SPITBF); // wait for the data to be sent out
      }

      // if got a 0, reset the encoder
      else if (read == 0) {
        QEI_reset(1);
        pos = QEI_read(1);
        SPI1BUF = pos;
        while (!SPI1STATbits.SPITBF); // wait for the data to be sent out
        //SPI1BUF = pos&0xFF;
        //while (!SPI1STATbits.SPITBF); // wait for the data to be sent out
      }
    }
  }
}
/*
void __attribute__((interrupt, auto_psv)) _SPI1Interrupt(void) {
  unsigned int buffer;

  // Retrieve data from receive buffer
  IFS0bits.SPI1IF = 0; //Clear the interrupt flag
  SPI1STATbits.SPIROV = 0; //Clear any errors
  buffer = SPI1BUF; //Read in SPI1 buffer

  if (buffer == 1) {
    write_SPI(QEI_read(1));
  } else if (buffer == 2) {
    QEI_reset(1);
    write_SPI(QEI_read(1));
  }
}
*/
void init_SPI(void) {
  // CN pins are shared, turn them off
  CNEN1 = 0;
  CNEN2 = 0;

  SPI1BUF = 0;
  IFS0bits.SPI1IF = 0; // Clear the Interrupt Flag
  IEC0bits.SPI1IE = 0; // Disable the Interrupt
  SPI1CON1bits.DISSCK = 0; // Internal Serial Clock is Enabled
  SPI1CON1bits.DISSDO = 0; // SDOx pin is controlled by the module
  SPI1CON1bits.SSEN = 1; // use the slave select pin
  SPI1CON1bits.MODE16 = 1; // Communication is 16 bit
  SPI1CON1bits.SMP = 0; // Input data is sampled at the middle of data
  SPI1CON1bits.CKE = 0; // Serial output data changes on transition
  SPI1CON1bits.CKP = 0; // Idle state for clock is a low level; active
  SPI1CON1bits.MSTEN = 0; // Master mode Disabled
  SPI1STATbits.SPIROV = 0; // No Receive Overflow has occurred
  SPI1STATbits.SPIEN = 1; // Enable SPI module
  // Interrupt Controller Settings
  //IFS0bits.SPI1IF = 0; // Clear the Interrupt Flag
  //IEC0bits.SPI1IE = 1; // Enable the Interrupt
}

void write_SPI(short command) {
  short temp;

  // slave select is automatic, comment for reference
  //PORTGbits.RG15 = 0; // lower the slave select line
  temp = SPI1BUF; // dummy read of the SPI1BUF register to clear the SPIRBF flag
  SPI1BUF = command; // write the data out to the SPI peripheral
  while (!SPI1STATbits.SPIRBF); // wait for the data to be sent out
  //PORTGbits.RG15 = 1; // raise the slave select line
}

unsigned int QEI_read(int QEI) {
  unsigned int QEI_Count = 0;

  // check CNTERR, but what if there is an error? what causes errors?
  if (QEI == 1 && QEI1CONbits.CNTERR == 0) {
    QEI_Count = POS1CNT; 
  }
  /*
  if (QEI == 2 && QEI2CONbits.CNTERR == 0) {
    QEI_Count = POS2CNT;
  }
  */
  return QEI_Count;
}

void QEI_enable(int QEI, int threshold) {
  if (threshold > 0xFFFF) {
    threshold = 0xFFFF;
  }
  if (threshold < 0) {
    threshold = 0;
  }
  
  if (QEI == 1 && !QEI1CONbits.CNTERR) {
    QEI1CONbits.QEIM = 0b111; // count in x4 mode
    MAX1CNT = threshold;
  }
  /*
  if (QEI == 2 && !QEI2CONbits.CNTERR) {
    QEI2CONbits.QEIM = 0b111;
    MAX2CNT = threshold;
  }
  */
}

void QEI_disable(int QEI) {
  if (QEI == 1) {
    QEI1CONbits.QEIM = 0; // don't count
  }
  /*
  if (QEI == 2) {
    QEI2CONbits.QEIM = 0;
  }
  */
}

void QEI_reset(int QEI) {
  if (QEI == 1) {
    POS1CNT = 32768; // return to half of max value
    QEI1CONbits.CNTERR = 0; // turn off any error
  }
  /*
  if (QEI == 2) {
    POS2CNT = 32768;
    QEI2CONbits.CNTERR = 0;
  }
  */
}




